"""Unit tests for the searchlight patch generator."""

import pytest
import numpy as np
from numpy.testing import assert_equal
from scipy import sparse

from mne_rsa import searchlight


class TestSearchLight:
    """Test the searchlight generator class."""

    def test_iter_spatio_temporal(self):
        """Test generating spatio-temporal searchlight patches."""
        dist = np.array([[0, 1, 2], [1, 0, 1], [2, 1, 0]])
        s = searchlight((10, 3, 4), dist, spatial_radius=2, temporal_radius=1)
        assert len(s) == 6
        assert s.shape == (3, 2)
        assert_equal(
            list(s),
            [
                (slice(None), np.array([0, 1]), slice(0, 3)),
                (slice(None), np.array([0, 1]), slice(1, 4)),
                (slice(None), np.array([0, 1, 2]), slice(0, 3)),
                (slice(None), np.array([0, 1, 2]), slice(1, 4)),
                (slice(None), np.array([1, 2]), slice(0, 3)),
                (slice(None), np.array([1, 2]), slice(1, 4)),
            ],
        )

        # Test using a sparse matrix
        dist_sparse = sparse.csr_matrix(dist)
        s = searchlight((10, 3, 4), dist_sparse, spatial_radius=2, temporal_radius=1)
        assert len(s) == 6
        assert s.shape == (3, 2)
        assert_equal(
            list(s),
            [
                (slice(None), np.array([0, 1]), slice(0, 3)),
                (slice(None), np.array([0, 1]), slice(1, 4)),
                (slice(None), np.array([0, 1, 2]), slice(0, 3)),
                (slice(None), np.array([0, 1, 2]), slice(1, 4)),
                (slice(None), np.array([1, 2]), slice(0, 3)),
                (slice(None), np.array([1, 2]), slice(1, 4)),
            ],
        )

        # Test using subsets of series and samples
        s = searchlight(
            (10, 3, 4),
            dist,
            spatial_radius=2,
            temporal_radius=1,
            sel_series=[1, 2],
            samples_from=2,
            samples_to=4,
        )
        assert len(s) == 2
        assert s.shape == (2, 1)
        assert_equal(
            list(s),
            [
                (slice(None), np.array([0, 1, 2]), slice(1, 4)),
                (slice(None), np.array([1, 2]), slice(1, 4)),
            ],
        )

    def test_iter_spatial(self):
        """Test generating spatial searchlight patches."""
        dist = np.array([[0, 1, 2], [1, 0, 1], [2, 1, 0]])

        s = searchlight((5, 10, 3, 4), dist, spatial_radius=2)
        assert len(s) == 3
        assert s.shape == (3,)
        assert_equal(
            list(s),
            [
                (slice(None), slice(None), np.array([0, 1]), slice(None)),
                (slice(None), slice(None), np.array([0, 1, 2]), slice(None)),
                (slice(None), slice(None), np.array([1, 2]), slice(None)),
            ],
        )

        s = searchlight((10, 3, 4), dist, spatial_radius=2)
        assert len(s) == 3
        assert s.shape == (3,)
        assert_equal(
            list(s),
            [
                (slice(None), np.array([0, 1]), slice(None)),
                (slice(None), np.array([0, 1, 2]), slice(None)),
                (slice(None), np.array([1, 2]), slice(None)),
            ],
        )

        s = searchlight((10, 3), dist, spatial_radius=2)
        assert len(s) == 3
        assert s.shape == (3,)
        assert_equal(
            list(s),
            [
                (slice(None), np.array([0, 1])),
                (slice(None), np.array([0, 1, 2])),
                (slice(None), np.array([1, 2])),
            ],
        )

        # Test using a sparse matrix
        dist_sparse = sparse.csr_matrix(dist)
        s = searchlight((10, 3), dist_sparse, spatial_radius=2)
        assert len(s) == 3
        assert s.shape == (3,)
        assert_equal(
            list(s),
            [
                (slice(None), np.array([0, 1])),
                (slice(None), np.array([0, 1, 2])),
                (slice(None), np.array([1, 2])),
            ],
        )

    def test_iter_temporal(self):
        """Test generating temporal searchlight patches."""
        s = searchlight((10, 3, 4), temporal_radius=1)
        assert len(s) == 2
        assert s.shape == (2,)
        assert_equal(
            list(s),
            [
                (slice(None), slice(None), slice(0, 3)),
                (slice(None), slice(None), slice(1, 4)),
            ],
        )

        s = searchlight((10, 4), temporal_radius=1)
        assert len(s) == 2
        assert s.shape == (2,)
        assert_equal(
            list(s),
            [
                (slice(None), slice(0, 3)),
                (slice(None), slice(1, 4)),
            ],
        )

    def test_iter_single(self):
        """Test generating a single searchlight patch."""
        s = searchlight((5, 10, 3, 4))
        assert len(s) == 1
        assert s.shape == tuple()
        assert_equal(next(s), tuple([slice(None)] * 4))

        s = searchlight((10, 3, 4))
        assert len(s) == 1
        assert s.shape == tuple()
        assert_equal(next(s), tuple([slice(None)] * 3))

        s = searchlight((10, 3))
        assert len(s) == 1
        assert s.shape == tuple()
        assert_equal(next(s), tuple([slice(None)] * 2))

        s = searchlight((10,))
        assert len(s) == 1
        assert s.shape == tuple()
        assert_equal(next(s), (slice(None),))

        # Test using subsets of series and samples
        s = searchlight((10, 3, 4), sel_series=[1, 2], samples_from=2, samples_to=4)
        assert len(s) == 1
        assert s.shape == tuple()
        assert_equal(next(s), (slice(None), [1, 2], slice(2, 4)))

    def test_invalid_input(self):
        """Test giving invalid input to searchlight generator."""
        dist = np.array([[0, 1, 2], [1, 0, 1], [2, 1, 0]])

        # Spatio-temporal case
        with pytest.raises(ValueError, match="no spatial dimension"):
            searchlight((10,), dist=dist, spatial_radius=1, temporal_radius=1)
        with pytest.raises(ValueError, match="no temporal dimension"):
            searchlight((10, 3), dist=dist, spatial_radius=1, temporal_radius=1)
        with pytest.raises(ValueError, match="no distance information"):
            searchlight((10, 3, 4), spatial_radius=1, temporal_radius=1)

        # Spatial case
        with pytest.raises(ValueError, match="no spatial dimension"):
            searchlight((10,), dist=dist, spatial_radius=1)
        with pytest.raises(ValueError, match="no distance information"):
            searchlight((10, 5), spatial_radius=1)
        with pytest.raises(ValueError, match="no spatial dimension"):
            searchlight((10,), sel_series=[1, 2])

        # Temporal case
        with pytest.raises(ValueError, match="no temporal dimension"):
            searchlight((10,), dist=dist, temporal_radius=1)
        with pytest.raises(ValueError, match="no temporal dimension"):
            searchlight((10,), dist=dist, samples_from=1, samples_to=3)



"""Functions concerning the creation of cross-validation folds."""

import numpy as np
from mne.utils import logger
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import OneHotEncoder


def create_folds(X, y=None, n_folds=None):
    """Group individual items into folds suitable for cross-validation.

    The ``y`` list should contain an integer label for each item in ``X``. Repetitions
    of the same item have the same integer label. Repeated items are distributed evenly
    across the folds, and averaged within a fold.

    Parameters
    ----------
    X : ndarray, shape (n_items, ...)
        For each item, all the features. The first dimension are the items and all other
        dimensions will be flattened and treated as features.
    y : ndarray of int, shape (n_items,) | None
        For each item, a number indicating the class to which the item belongs. When
        ``None``, each item is assumed to belong to a different class. Defaults to
        ``None``.
    n_folds : int | sklearn.BaseCrossValidator | None
        Number of cross-validation folds to use when computing the distance metric.
        Folds are created based on the ``y`` parameter. Specify ``None`` to use the
        maximum number of folds possible, given the data. Alternatively, you can pass a
        Scikit-Learn cross validator object (e.g. ``sklearn.model_selection.KFold``) to
        assert fine-grained control over how folds are created. Defaults to ``None``.

    Returns
    -------
    folds : ndarray, shape (n_folds, n_items, ...)
        The folded data.

    """
    if y is None:
        # No folding
        return X[np.newaxis, ...]

    y = np.asarray(y)
    if len(y) != len(X):
        raise ValueError(
            f"The length of y ({len(y)}) does not match the "
            f"number of items ({len(X)})."
        )

    y_one_hot = _convert_to_one_hot(y)
    n_items = y_one_hot.shape[1]

    if n_folds is None:
        # Set n_folds to maximum value
        n_folds = len(X) // n_items
        logger.info(
            f"Automatic dermination of folds: {n_folds}" + " (no cross-validation)"
            if n_folds == 1
            else ""
        )

    if n_folds == 1:
        # Making one fold is easy
        folds = [_compute_item_means(X, y_one_hot)]
    elif hasattr(n_folds, "split"):
        # Scikit-learn object passed as `n_folds`
        folds = []
        for _, fold in n_folds.split(X, y):
            folds.append(_compute_item_means(X, y_one_hot, fold))
    else:
        # Use StratifiedKFold as folding strategy
        folds = []
        for _, fold in StratifiedKFold(n_folds).split(X, y):
            folds.append(_compute_item_means(X, y_one_hot, fold))
    return np.array(folds)


def _convert_to_one_hot(y):
    """Convert the labels in y to one-hot encoding."""
    y = np.asarray(y)
    if y.ndim == 1:
        y = y[:, np.newaxis]

    if y.ndim == 2 and y.shape[1] == 1:
        # y needs to be converted
        enc = OneHotEncoder(categories="auto").fit(y)
        return enc.transform(y).toarray()
    elif y.ndim > 2:
        raise ValueError("Wrong number of dimensions for `y`.")
    else:
        # y is probably already in one-hot form. We're not going to test this
        # explicitly, as it would take too long.
        return y


def _compute_item_means(X, y_one_hot, fold=slice(None)):
    """Compute the mean data for each item inside a fold."""
    X = X[fold]
    y_one_hot = y_one_hot[fold]
    n_per_class = y_one_hot.sum(axis=0)

    # The following computations go much faster when X is flattened.
    orig_shape = X.shape
    X_flat = X.reshape(len(X), -1)

    # Compute the mean for each item using matrix multiplication
    means = (y_one_hot.T @ X_flat) / n_per_class[:, np.newaxis]

    # Undo the flattening of X
    return means.reshape((len(means),) + orig_shape[1:])



"""Classes and functions having to do with creating searchlights."""

import numpy as np
from mne.utils import logger


class searchlight:
    """Generate indices for searchlight patches.

    Generates a sequence of tuples that can be used to index a data array.
    Depending on the spatial and temporal radius, each tuple extracts a
    searchlight patch along time, space or both.

    This function is flexible in regards to shape of the data array. The
    intepretation of the dimensions is as follows:

    4 or more dimensions
        ``(n_folds, n_items, n_series, n_samples, ...)``

    3 dimensions
        ``(n_items, n_series, n_samples)``

    2 dimensions
        ``(n_items, n_series)`` when ``spatial_radius`` is not ``None``.
        ``(n_items, n_samples)`` when ``temporal_radius`` is not ``None``.

    1 dimension
        ``(n_items,)``

    The returned tuples will match the dimensions of the data array.

    Parameters
    ----------
    shape : tuple of int
        The shape of the data array to compute the searchlight patches for, as
        obtained with the ``.shape`` attribute.
    dist : ndarray or sparse matrix, shape (n_series, n_series) | None
        The distances between all source points or sensors in meters.
        This parameter needs to be specified if a ``spatial_radius`` is set.

        Since the distance matrix can be huge, sparse matrices are also
        supported. When the distance matrix is sparse, all zero distances are
        treated as infinity. This allows you to skip far away points during
        your distance computations.

        Defaults to ``None``.
    spatial_radius : float | list of list of int | None
        This controls how spatial patches will be created. There are several
        ways to do this:

        The first way is to specify a spatial radius in meters. In this case,
        the ``dist`` parameter must also be specified. This will create a
        searchlight where each patch contains all source points within this
        radius.

        The second way is to specify a list of predefined patches. In this
        case, each element of the list should itself be a list of integer
        indexes along the spatial dimension of the data array. Each element of
        this list will become a separate patch using the data at the specified
        indices.

        The third way is to set this to ``None``, which will disable the making
        of spatial patches and only perform the searchlight over time. This can
        be thought of as pooling everything into a single spatial patch.

        Defaults to``None``.
    temporal_radius : int | None
        The temporal radius of the searchlight patch in samples. Set to
        ``None`` to only perform the searchlight over sensors/source points.
        Defaults to ``None``.
    sel_series : ndarray, shape (n_selected_series,) | None
        When set, searchlight patches will only be generated for the subset of
        time series with the given indices. Defaults to ``None``, in which case
        patches for all series are generated.
    samples_from : int
        When set, searchlight patches will only be generated for the subset of
        time samples with indices equal or greater than the given value. Only
        used when the given data shape includes a temporal dimension.
        Defaults to 0.
    samples_to : int
        When set, searchlight patches will only be generated for the subset of
        time samples with indices up to, but not including, the given value.
        Only used when the given data shape includes a temporal dimension.
        Defaults to -1, which means there is no upper bound.

    Yields
    ------
    patch : tuple of (slice | ndarray)
        A single searchlight patch. Each element of the tuple corresponds to a
        dimension of the data array and can be used to index along this
        dimension to extract the searchlight patch.

    Attributes
    ----------
    shape

    """

    def __init__(
        self,
        shape,
        dist=None,
        spatial_radius=None,
        temporal_radius=None,
        sel_series=None,
        samples_from=0,
        samples_to=-1,
    ):
        # Interpret the dimensions of the data array (see docstring)
        n_dims = len(shape)
        if n_dims >= 4:
            self.series_dim = 2
            self.samples_dim = 3
        elif n_dims == 3:
            self.series_dim = 1
            self.samples_dim = 2
        elif n_dims == 2 and spatial_radius is not None:
            self.series_dim = 1
            self.samples_dim = None
        elif n_dims == 2 and temporal_radius is not None:
            self.series_dim = None
            self.samples_dim = 1
        else:
            self.series_dim = None
            self.samples_dim = None

        self.dist = dist
        self.spatial_radius = spatial_radius
        self.temporal_radius = temporal_radius
        self.sel_series = sel_series

        # Boundry checking for samples_from and samples_to. Only relevant if
        # there is a temporal dimension to the data.
        if samples_from != 0 or samples_to != -1:
            if self.samples_dim is None:
                raise ValueError(
                    "Cannot select samples:"
                    f"the provided data shape {shape} has no "
                    "temporal dimension."
                )
            n_samples = shape[self.samples_dim]
            if samples_from < 0 or samples_from > n_samples:
                raise ValueError(
                    f"`samples_from={samples_from}` is out "
                    f"of bounds given data shape ({shape})."
                )
            if samples_to > n_samples:
                raise ValueError(
                    f"`samples_to={samples_to}` is out "
                    f"of bounds given data shape ({shape})."
                )
            if samples_to != -1 and samples_to < samples_from:
                raise ValueError(
                    f"`samples_to={samples_to} is smaller "
                    f"than `samples_from={samples_from}."
                )
        self.samples_from = samples_from
        self.samples_to = samples_to

        # Will we be creating spatial searchlight patches?
        if self.spatial_radius is not None:
            if self.series_dim is None:
                raise ValueError(
                    "Cannot create spatial searchlight patches: "
                    f"the provided data shape ({shape}) has no "
                    "spatial dimension."
                )
            # If spatial radius is a number, we will be making searchlight
            # patches based on distance computations. Alternatively, a list of
            # predefined spatial patches may be provided, and we don't need
            # `dist`.
            if type(self.spatial_radius) in [float, int]:
                if self.dist is None:
                    raise ValueError(
                        "A spatial radius was requested, but no "
                        "distance information was specified "
                        "(=dist parameter)."
                    )
                # Compressed Sparse Row format is optimal for our computations
                from scipy.sparse import issparse

                if issparse(self.dist):
                    self.dist = self.dist.tocsr()
                if self.sel_series is None:
                    self.sel_series = np.arange(shape[self.series_dim])
            else:
                # Explicit spatial patches were provided
                if self.sel_series is None:
                    self.sel_series = np.arange(len(self.spatial_radius))

        # Will we be creating temporal searchlight patches?
        if temporal_radius is not None:
            if self.samples_dim is None:
                raise ValueError(
                    "Cannot create temporal searchlight patches: "
                    f"the provided data shape ({shape}) has no "
                    "temporal dimension."
                )
            n_samples = shape[self.samples_dim]

            # Compute the centers of the searchlight patches in time. Make sure
            # that adding/subtracting the temporal_radius does not produce
            # array out of bounds errors.
            samples_min = temporal_radius
            samples_max = n_samples - temporal_radius
            if samples_min > samples_max:
                raise ValueError(
                    f"Temporal radius ({temporal_radius}) too large for the "
                    f"given data shape ({shape})."
                )
            self.time_centers = list(
                range(
                    np.clip(samples_from, samples_min, samples_max),
                    np.clip(
                        n_samples if samples_to == -1 else samples_to,
                        samples_min,
                        samples_max,
                    ),
                )
            )

        # Create a template for the patches that will be generated that is
        # compatible with the data array dimensions. By default, we select
        # everything along every dimension, taking `sel_series`, `samples_from`
        # and `samples_to` into account. This template will be filled-in inside
        # the __iter__ function.
        self.patch_template = [slice(None)] * n_dims
        if self.sel_series is not None:
            if self.series_dim is None:
                raise ValueError(
                    "Cannot select series:"
                    f"the provided data shape {shape} has no "
                    "spatial dimension."
                )
            self.patch_template[self.series_dim] = self.sel_series
        if self.samples_from != 0 or self.samples_to != -1:
            if self.samples_dim is None:
                raise ValueError(
                    "Cannot select samples:"
                    f"the provided data shape {shape} has no "
                    "temporal dimension."
                )
            self.patch_template[self.samples_dim] = slice(
                self.samples_from, self.samples_to
            )

        # Setup the main generator function that will be providing the
        # searchlight patches.
        self._generator = iter(self)

    def __iter__(self):
        """Get an iterator over the searchlight patches."""
        if self.spatial_radius is not None and self.temporal_radius is not None:
            iterator = self._iter_spatio_temporal()
        elif self.spatial_radius is not None:
            iterator = self._iter_spatial()
        elif self.temporal_radius is not None:
            iterator = self._iter_temporal()
        else:
            # Single searchlight patch only
            iterator = iter([tuple(self.patch_template)])
        return iterator

    def __next__(self):
        """Generate searchlight patches."""
        return next(self._generator)

    def _iter_spatio_temporal(self):
        """Generate spatio-temporal searchlight patches."""
        logger.info("Creating spatio-temporal searchlight patches")
        patch = list(self.patch_template)  # Copy the template
        for series in self.sel_series:
            # Compute all spatial locations in the searchligh path.
            if type(self.spatial_radius) in [float, int]:
                spat_ind = _get_in_radius(self.dist, series, self.spatial_radius)
            else:
                spat_ind = self.spatial_radius[series]
            patch[self.series_dim] = spat_ind
            for sample in self.time_centers:
                temp_ind = slice(
                    sample - self.temporal_radius, sample + self.temporal_radius + 1
                )
                patch[self.samples_dim] = temp_ind
                yield tuple(patch)

    def _iter_spatial(self):
        """Generate spatial searchlight patches only."""
        logger.info("Creating spatial searchlight patches")
        patch = list(self.patch_template)  # Copy the template
        for series in self.sel_series:
            if type(self.spatial_radius) in [float, int]:
                spat_ind = _get_in_radius(self.dist, series, self.spatial_radius)
            else:
                spat_ind = self.spatial_radius[series]
            patch[self.series_dim] = spat_ind
            yield tuple(patch)

    def _iter_temporal(self):
        """Generate temporal searchlight patches only."""
        logger.info("Creating temporal searchlight patches")
        patch = list(self.patch_template)  # Copy the template
        for sample in self.time_centers:
            patch[self.samples_dim] = slice(
                sample - self.temporal_radius, sample + self.temporal_radius + 1
            )
            yield tuple(patch)

    @property
    def shape(self):
        """Get the number of generated patches along multiple dimensions.

        This is useful for re-shaping the result obtained after consuming the
        this generator.

        Returns
        -------
        shape : tuple of int
            For a spatio-temporal searchlight:
                Two elements: the number of time-series and number of time
                samples for which patches are generated.
            For a spatial searchlight:
                One element: the number of time-series for which patches are
                generated.
            For a temporal searchlight:
                One element: the number of time-samples for which patches are
                generated.
            For no searchlight:
                Zero elements.

        """
        if self.spatial_radius is not None and self.temporal_radius is not None:
            return (len(self.sel_series), len(self.time_centers))
        elif self.spatial_radius is not None:
            return (len(self.sel_series),)
        elif self.temporal_radius is not None:
            return (len(self.time_centers),)
        else:
            return tuple()

    def __len__(self):
        """Get total number of searchlight patches that will be generated."""
        total = 1
        for n in self.shape:  # Number of patches generated in each dimension
            total *= n
        return total


def _get_in_radius(dist, seed, radius):
    """Obtain indices for all points within the given radius from a seed point.

    Takes care to work with sparse matrices too.

    Parameters
    ----------
    dist : ndarray or sparse matrix, shape (n_points, n_points)
        The distances between all points.
    seed : int
        The index of the point used as a seed.
    radius : float
        The maximum distance that points can be to be included.

    Returns
    -------
    ind : ndarray, shape (n_points_in_radius,)
        Indices of all points in the given radius from the seed point.

    """
    from scipy.sparse import issparse

    if issparse(dist):
        # Treat all zero distances as missing data
        ind = dist[seed].nonzero()[1]
        # Find indices for points within the radius
        ind = ind[dist[seed].data < radius]
        ind.sort()
        # Be sure to add the seed point, which has distance of 0 to itself
        ind = np.hstack((ind, [seed]))  # Di
    else:
        ind = np.flatnonzero(dist[seed] < radius)
    return sorted(ind)



"""Functions related to visualization of RDMs."""

import types
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
from mne.viz import Brain
from mne.viz.topo import _iter_topography
from scipy.spatial import distance


def plot_rdms(rdms, names=None, items=None, n_rows=1, cmap="viridis", title=None):
    """Plot one or more RDMs.

    Parameters
    ----------
    rdms : ndarray | list of ndarray
        The RDM or list of RDMs to plot. The RDMs can either be two-dimensional (n_items
        x n_items) matrices or be in condensed form.
    names : str | list of str | None
        For each given RDM, a name to show above it. Defaults to no names.
    items : list of str | None
        The each item (row/col) in the RDM, a string description. This will be displayed
        along the axes. Defaults to None which means the items will be numbered.
    n_rows : int
        Number of rows to use when plotting multiple RDMs at once. Defaults to 1.
    cmap : str
        Matplotlib colormap to use. See
        https://matplotlib.org/gallery/color/colormap_reference.html
        for all possibilities. Defaults to 'viridis'.
    title : str | None
        Title for the entire figure. Defaults to no title.

    Returns
    -------
    fig : matplotlib figure
        The figure produced by matplotlib

    """
    if not isinstance(rdms, list):
        rdms = [rdms]

    if isinstance(names, str):
        names = [names]
    if names is not None and len(names) != len(rdms):
        raise ValueError(
            f"Number of given names ({len(names)}) does not "
            f"match the number of RDMs ({len(rdms)})"
        )

    n_cols = int(np.ceil(len(rdms) / n_rows))
    fig = plt.figure(figsize=(2 * n_cols, 2 * n_rows))

    ax = fig.subplots(n_rows, n_cols, sharex=True, sharey=True, squeeze=False)
    for row in range(n_rows):
        for col in range(n_cols):
            i = row * n_cols + col % n_cols
            if i < len(rdms):
                rdm = rdms[i]
                if rdm.ndim == 1:
                    rdm = distance.squareform(rdm)
                elif rdm.ndim > 2:
                    raise ValueError(f"Invalid shape {rdm.shape} for RDM")
                im = ax[row, col].imshow(rdm, cmap=cmap)

                if names is not None:
                    name = names[i]
                    ax[row, col].set_title(name)
                if items is not None:
                    ax[row, col].set_xticks(np.arange(len(items)))
                    ax[row, col].set_xticklabels(items)
                    ax[row, col].set_yticks(np.arange(len(items)))
                    ax[row, col].set_yticklabels(items)
            else:
                ax[row, col].set_visible(False)

    plt.colorbar(im, ax=ax)
    if title is not None:
        plt.suptitle(title)
    return fig


def _click_func(ax, ch_idx, rdms, cmap):
    """Plot a single RDM interactively.

    Parameters
    ----------
    ax: matplotlib.Axes.axes
        Axes.axes object on which a new single RDM is plotted.
    ch_idx: int
        Index of a channel.
    rdms: ndarray, shape (n_sensors, n_rdm_datapoint)
        RDMs of MEG recordings; there's one RDM for each sensor.
    cmap: str
        Colormap used for plotting RDMs. Check ``matplotlib.pyplot.imshow`` for details.

    """
    rdm = rdms[ch_idx]
    rdm = distance.squareform(rdm)
    ax.imshow(rdm, cmap=cmap)


def _plot_rdms_topo_timepoint(
    rdms,
    info,
    layout=None,
    fig=None,
    title=None,
    axis_facecolor="w",
    axis_spinecolor="w",
    fig_facecolor="w",
    figsize=(6.4, 4.8),
    cmap="viridis",
    show=False,
):
    """Plot RDMs on 2D MEG topography.

    Parameters
    ----------
    rdms: ndarray, shape (n_sensors, n_rdm_datapoints) | generator
        RDMs of MEG recordings; one RDM for each sensor. Can also be a generator of RDMs
        as produced by the :func:`rdm_epochs`,
        :func:`rdm_evokeds` or :func:`rdm_array` functions.
    info: mne.io.meas_info.Info
        Info object that contains meta data of MEG recordings.
    layout: mne.channels.layout.Layout | None
        Layout objects containing sensor layout info. The default (``None``) will figure
        out layout based on info.
    fig: matplotlib.pyplot.Figure | None
        Figure object on which RDMs on 2D MEG topography are plotted. The default
        (``None``) creates a new Figure object.
    title: str | None
        Title of the plot, used only when ``fig=None``. The default (``None``) puts no
        title in the figure.
    axis_facecolor: str
        Face color of the each RDM. Defaults to 'w', white.
    axis_spinecolor: str
        Spine color of each RDM. Defaults to 'w', white.
    fig_facecolor: str
        Face color of the entire topography. Defaults to 'w', white.
    figsize: tuple of float
        Figure size. The first element specify width and the second height.
        Defaults to (6.4, 4.8).
    cmap: str
        Colormap used for plotting RDMs. Defaults to 'viridis'.
        Check :func:`matplotlib.pyplot.imshow` for details.
    show: bool
        Whether to display the generated figure. Defaults to False.

    Returns
    -------
    fig: matplotlib.pyplot.Figure
        Figure object in which RDMs are plotted on 2D MEG topography.

    """
    on_pick = partial(_click_func, rdms=rdms, cmap=cmap)

    if fig is None:
        fig = plt.figure(figsize=figsize)
        if title is not None:
            fig.suptitle(title, x=0.98, horizontalalignment="right")
    else:
        fig = plt.figure(fig.number)

    my_topo_plot = _iter_topography(
        info=info,
        layout=layout,
        on_pick=on_pick,
        fig=fig,
        axis_facecolor=axis_facecolor,
        axis_spinecolor=axis_spinecolor,
        fig_facecolor=fig_facecolor,
        unified=False,
    )

    for i, (ax, _) in enumerate(my_topo_plot):
        rdms_i = rdms[i]
        rdms_i = distance.squareform(rdms_i)
        ax.imshow(rdms_i, cmap=cmap)

    if show:
        fig.show()

    return fig


def plot_rdms_topo(
    rdms,
    info,
    time=None,
    layout=None,
    fig=None,
    axis_facecolor="w",
    axis_spinecolor="w",
    fig_facecolor="w",
    figsize=(6.4, 4.8),
    cmap="viridis",
    show=True,
):
    """Plot RDMs on 2D sensor topography.

    Parameters
    ----------
    rdms: ndarray | numpy.memmap, shape (n_sensors,[ n_times,] n_rdm_datapts)
        RDMs of MEG/EEG recordings; one RDM for each sensor and time point.
    info: mne.io.meas_info.Info
        Info object that contains meta data of MEG/EEG recordings.
    time: int | [int, int] | None
        A time point (int) or time window ([int, int]) for which RDMs are plotted. When
        a time window is given, averge RDMs for the window are plotted. The default
        (``None``) plots the average RDMs of all the time points. Start of the time
        window is inclusive, while the end is exclusive.
    layout: mne.channels.layout.Layout, optional
        Layout objects containing sensor layout info. The default, ``layout=None``, will
        figure out layout based on info.
    fig: matplotlib.pyplot.Figure | None, optional
        Figure object on which RDMs on 2D sensor topography are plotted. The default
        (``None``) creates a new Figure object with a title based on time parameter.
    axis_facecolor: str, optional
        Face color of the each RDM. Defaults to 'w', white.
    axis_spinecolor: str, optional
        Spine color of each RDM. Defaults to 'w', white.
    fig_facecolor: str, optional
        Face color of the entire topography. Defaults to 'w', white.
    figsize: tuple of float, optional
        Figure size. The first element specify width and the second height.
        Defaults to (6.4, 4.8).
    cmap: str, optional
        Colormap used for plotting RDMs. Defaults to 'viridis'.
        Check :func:`matplotlib.pyplot.imshow` for details.
    show: bool, optional
        Whether to display the generated figure. Defaults to ``True``.

    Returns
    -------
    fig: matplotlib.pyplot.Figure
        Figure object in which RDMs are plotted on 2D sensor topography.

    """
    if isinstance(rdms, types.GeneratorType):
        rdms = np.array(list(rdms))

    if rdms.ndim != 2 and rdms.ndim != 3:
        raise ValueError(
            "rdms have to be a 2D or 3D ndarray or numpy.memmap, "
            "[n_sensors,[ n_times,] n_rdm_datapoints]"
        )
    if len(rdms.shape) == 2:
        rdms = rdms[:, np.newaxis, :]
    if time is None:
        time = [0, rdms.shape[1]]
    if isinstance(time, int):
        time = [time, time + 1]
    if not isinstance(time, list):
        raise TypeError("time has to be int, list of [int, int] or None.")
    if (not all(isinstance(i, int) for i in time)) or (len(time) != 2):
        raise TypeError("time has to be int, list of [int, int] or None.")
    if time[0] >= time[1]:
        raise ValueError(
            "The start of the time window has to be smaller "
            "than the end of the time window."
        )
    if time[0] < 0 or time[1] > rdms.shape[1]:
        raise ValueError(
            "The time window is out of range. The minimum is 0 "
            f"and the maximum is {rdms.shape[1]}"
        )
    if (fig is not None) and (not isinstance(fig, plt.Figure)):
        raise TypeError("fig has to be matplotlib.pyplot.Figure or None.")

    rdms_cropped = rdms[:, time[0] : time[1], :]
    rdms_avg = rdms_cropped.mean(axis=1)
    # set title to time window
    if time[0] + 1 != time[1]:
        title = f"From {time[0]} (inclusive) to {time[1]} (exclusive)"
    else:
        title = f"Time point: {time[0]}"

    fig = _plot_rdms_topo_timepoint(
        rdms_avg,
        info,
        fig=fig,
        layout=layout,
        title=title,
        axis_facecolor=axis_facecolor,
        axis_spinecolor=axis_spinecolor,
        fig_facecolor=fig_facecolor,
        figsize=figsize,
        cmap=cmap,
        show=show,
    )
    return fig


def plot_roi_map(values, rois, subject, subjects_dir, cmap="plasma", alpha=1.0):
    """Plot ROI values on a FreeSurfer brain.

    Parameters
    ----------
    values : array-like, shape (n_rois,)
        The values to plot. One value per ROI.
    rois : list of mne.Label
        The labels corrsponding to the ROIs.
    subject : str
        The name of the FreeSurfer subject to plot the brain for.
    subjects_dir : str
        The folder in which the FreeSurfer subject data is kept. Inside this folder
        should be a folder with the same name as the `subject` parameter.
    cmap : str
        The name of the matplotlib colormap to use. Defaults to 'plasma'.
    alpha : float
        The alpha (opacity, 1.0 is fully opaque, 0.0 is fully transparant) of the data
        being plotted on top of the brain.

    Returns
    -------
    brain : mne.viz.Brain
        The MNE-Python brain plotting object that was created and currently
        being shown. You can use this to modify the plot.

    """
    cmap = plt.get_cmap(cmap)
    max_val = np.max(values)
    brain = Brain(
        subject=subject, subjects_dir=subjects_dir, surf="inflated", hemi="both"
    )
    labels_lh = np.zeros(len(brain.geo["lh"].coords), dtype=int)
    labels_rh = np.zeros(len(brain.geo["rh"].coords), dtype=int)
    ctab_lh = list()
    ctab_rh = list()
    for i, (roi, value) in enumerate(zip(rois, values), 1):
        if roi.hemi == "lh":
            labels = labels_lh
            ctab = ctab_lh
        else:
            labels = labels_rh
            ctab = ctab_rh
        labels[roi.vertices] = i
        ctab.append([int(x * 255) for x in cmap(value / max_val)[:4]] + [i])
    ctab_lh = np.array(ctab_lh)
    ctab_rh = np.array(ctab_rh)
    brain.add_annotation(
        [(labels_lh, ctab_lh), (labels_rh, ctab_rh)], borders=False, alpha=alpha
    )
    return brain



# encoding: utf-8
"""Methods to compute representational similarity analysis (RSA)."""

import numpy as np
from joblib import Parallel, delayed
from scipy import stats

from .folds import create_folds
from .rdm import _ensure_condensed, compute_rdm, compute_rdm_cv
from .searchlight import searchlight

try:
    # Version 1.8.0 and up
    from scipy.stats._stats_py import _kendall_dis
except ImportError:
    from scipy.stats.stats import _kendall_dis


def _kendall_tau_a(x, y):
    """Compute Kendall's Tau metric, A-variant.

    Taken from scipy.stats.kendalltau and modified to be the tau-a variant.
    """
    x = np.asarray(x).ravel()
    y = np.asarray(y).ravel()

    if x.size != y.size:
        raise ValueError(
            "All inputs to `kendalltau` must be of the same size,"
            " found x-size %s and y-size %s" % (x.size, y.size)
        )
    elif not x.size or not y.size:
        return np.nan  # Return NaN if arrays are empty

    def count_rank_tie(ranks):
        cnt = np.bincount(ranks).astype("int64", copy=False)
        cnt = cnt[cnt > 1]
        return (
            (cnt * (cnt - 1) // 2).sum(),
            (cnt * (cnt - 1.0) * (cnt - 2)).sum(),
            (cnt * (cnt - 1.0) * (2 * cnt + 5)).sum(),
        )

    size = x.size
    perm = np.argsort(y)  # sort on y and convert y to dense ranks
    x, y = x[perm], y[perm]
    y = np.r_[True, y[1:] != y[:-1]].cumsum(dtype="intp")

    # stable sort on x and convert x to dense ranks
    perm = np.argsort(x, kind="mergesort")
    x, y = x[perm], y[perm]
    x = np.r_[True, x[1:] != x[:-1]].cumsum(dtype="intp")

    dis = _kendall_dis(x, y)  # discordant pairs

    obs = np.r_[True, (x[1:] != x[:-1]) | (y[1:] != y[:-1]), True]
    cnt = np.diff(np.nonzero(obs)[0]).astype("int64", copy=False)

    ntie = (cnt * (cnt - 1) // 2).sum()  # joint ties
    xtie, x0, x1 = count_rank_tie(x)  # ties in x, stats
    ytie, y0, y1 = count_rank_tie(y)  # ties in y, stats

    tot = (size * (size - 1)) // 2

    if xtie == tot or ytie == tot:
        return np.nan

    # Note that tot = con + dis + (xtie - ntie) + (ytie - ntie) + ntie
    #               = con + dis + xtie + ytie - ntie
    con_minus_dis = tot - xtie - ytie + ntie - 2 * dis
    tau = con_minus_dis / tot
    # Limit range to fix computational errors
    tau = min(1.0, max(-1.0, tau))

    return tau


def _consolidate_masks(masks):
    if type(masks[0]) == slice:
        mask = slice(None)
    else:
        mask = masks[0]
        for m in masks[1:]:
            mask &= m
    return mask


def _partial_correlation(rdm_data, rdm_model, masks=None, type="pearson"):
    """Compute partial Pearson/Spearman correlation."""
    if len(rdm_model) == 1:
        raise ValueError(
            "Need more than one model RDM to use partial " "correlation as metric."
        )
    if type not in ["pearson", "spearman"]:
        raise ValueError("Correlation type must be either 'pearson' or " "'spearman'")

    if masks is not None:
        mask = _consolidate_masks(masks)
        rdm_model = [rdm[mask] for rdm in rdm_model]
        rdm_data = rdm_data[mask]

    X = np.vstack([rdm_data] + rdm_model).T
    if type == "spearman":
        X = np.apply_along_axis(stats.rankdata, 0, X)
    X = X - X.mean(axis=0)
    cov_X_inv = np.linalg.pinv(X.T @ X)
    norm = np.sqrt(np.outer(np.diag(cov_X_inv), np.diag(cov_X_inv)))
    R_partial = cov_X_inv / norm
    return -R_partial[0, 1:]


def rsa_gen(rdm_data_gen, rdm_model, metric="spearman", ignore_nan=False):
    """Generate RSA values between data and model RDMs.

    Will yield RSA scores for each data RDM.

    Parameters
    ----------
    rdm_data_gen : generator of ndarray, shape (n_items, n_items)
        The generator for data RDMs
    rdm_model : ndarray, shape (n_items, n_items) | list of ndarray
        The model RDM, or list of model RDMs.
    metric : str
        The RSA metric to use to compare the RDMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    ignore_nan : bool
        Whether to treat NaN's as missing values and ignore them when computing the
        distance metric. Defaults to ``False``.

        .. versionadded:: 0.8

    Yields
    ------
    rsa_val : float | ndarray, shape (len(rdm_model),)
        For each data RDM, the representational similarity with the model RDM. When
        multiple model RDMs are specified, this will be a 1D array of similarities,
        comparing the data RDM with each model RDM.

    See Also
    --------
    rsa

    """
    if isinstance(rdm_model, list):
        return_array = True
        rdm_model = [_ensure_condensed(rdm, "rdm_model") for rdm in rdm_model]
    else:
        return_array = False
        rdm_model = [_ensure_condensed(rdm_model, "rdm_model")]

    if ignore_nan:
        masks = [~np.isnan(rdm) for rdm in rdm_model]
    else:
        masks = [slice(None)] * len(rdm_model)

    for rdm_data in rdm_data_gen:
        rdm_data = _ensure_condensed(rdm_data, "rdm_data")
        if ignore_nan:
            data_mask = ~np.isnan(rdm_data)
            masks = [m & data_mask for m in masks]
        rsa_vals = _rsa_single_rdm(rdm_data, rdm_model, metric, masks)
        if return_array:
            yield np.asarray(rsa_vals)
        else:
            yield rsa_vals[0]


def _rsa_single_rdm(rdm_data, rdm_model, metric, masks):
    """Compute RSA between a single data RDM and one or more model RDMs."""
    if metric == "spearman":
        rsa_vals = [
            stats.spearmanr(rdm_data[mask], rdm_model_[mask])[0]
            for rdm_model_, mask in zip(rdm_model, masks)
        ]
    elif metric == "pearson":
        rsa_vals = [
            stats.pearsonr(rdm_data[mask], rdm_model_[mask])[0]
            for rdm_model_, mask in zip(rdm_model, masks)
        ]
    elif metric == "kendall-tau-a":
        rsa_vals = [
            _kendall_tau_a(rdm_data[mask], rdm_model_[mask])
            for rdm_model_, mask in zip(rdm_model, masks)
        ]
    elif metric == "partial":
        rsa_vals = _partial_correlation(rdm_data, rdm_model, masks)
    elif metric == "partial-spearman":
        rsa_vals = _partial_correlation(rdm_data, rdm_model, masks, type="spearman")
    elif metric == "regression":
        mask = _consolidate_masks(masks)
        rdm_model = [rdm[mask] for rdm in rdm_model]
        rdm_data = rdm_data[mask]
        X = np.atleast_2d(np.array(rdm_model)).T
        X = X - X.mean(axis=0)
        y = rdm_data - rdm_data.mean()
        rsa_vals = np.linalg.lstsq(X, y, rcond=None)[0]
    else:
        raise ValueError(
            "Invalid RSA metric, must be one of: 'spearman', "
            "'pearson', 'partial', 'partial-spearman', "
            "'regression' or 'kendall-tau-a'."
        )
    return rsa_vals


def rsa(
    rdm_data,
    rdm_model,
    metric="spearman",
    ignore_nan=False,
    n_data_rdms=None,
    n_jobs=1,
    verbose=False,
):
    """Perform RSA between data and model RDMs.

    Parameters
    ----------
    rdm_data : ndarray, shape (n_items, n_items) | list | generator
        The data RDM (or list/generator of data RDMs).
    rdm_model : ndarray, shape (n_items, n_items) | list of ndarray
        The model RDM (or list of model RDMs).
    metric : str
        The RSA metric to use to compare the RDMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    ignore_nan : bool
        Whether to treat NaN's as missing values and ignore them when computing
        the distance metric. Defaults to ``False``.

        .. versionadded:: 0.8
    n_data_rdms : int | None
        The number of data RDMs. This is useful when displaying a progress bar, so an
        estimate can be made of the computation time remaining. This information is
        available if ``rdm_data`` is an array or a list, but if it is a generator, this
        information is not available and you may want to set it explicitly.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to use all
        available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need the tqdm
        python module installed. Defaults to False.

    Returns
    -------
    rsa_val : float | ndarray, shape (len(rdm_data), len(rdm_model))
        Depending on whether one or more data and model RDMs were specified, a single
        similarity value or a 2D array of similarity values for each data RDM versus
        each model RDM.

    See Also
    --------
    rsa_gen

    """
    return_array = False
    if isinstance(rdm_data, list) or hasattr(rdm_data, "__next__"):
        return_array = True
    else:
        rdm_data = [rdm_data]

    if verbose:
        from tqdm import tqdm

        if n_data_rdms is not None:
            total = n_data_rdms
        elif hasattr(rdm_data, "__len__"):
            total = len(rdm_data)
        else:
            total = None
        rdm_data = tqdm(rdm_data, total=total, unit="RDM")

    if n_jobs == 1:
        rsa_vals = list(rsa_gen(rdm_data, rdm_model, metric, ignore_nan))
    else:

        def process_single_rdm(rdm):
            return next(rsa_gen([rdm], rdm_model, metric, ignore_nan))

        rsa_vals = Parallel(n_jobs)(
            delayed(process_single_rdm)(rdm) for rdm in rdm_data
        )
    if return_array:
        return np.asarray(rsa_vals)
    else:
        return rsa_vals[0]


def rsa_array(
    X,
    rdm_model,
    patches=None,
    data_rdm_metric="correlation",
    data_rdm_params=dict(),
    rsa_metric="spearman",
    ignore_nan=False,
    y=None,
    n_folds=1,
    n_jobs=1,
    verbose=False,
):
    """Perform RSA on an array of data, possibly in a searchlight pattern.

    Parameters
    ----------
    X : ndarray, shape (n_items, n_series, n_times)
        An array containing the data.
    rdm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model RDM, see :func:`compute_rdm`. For efficiency, you can give it in
        condensed form, meaning only the upper triangle of the matrix as a vector. See
        :func:`scipy.spatial.distance.squareform`. To perform RSA against multiple
        models at the same time, supply a list of model RDMs.

        Use :func:`compute_rdm` to compute RDMs.
    patches : generator of tuples | None
        Searchlight patches as generated by :class:`searchlight`. If ``None``, no
        searchlight is used. Defaults to ``None``.
    data_rdm_metric : str
        The metric to use to compute the data RDMs. This can be any metric supported by
        the scipy.distance.pdist function. Defaults to 'correlation'.
    data_rdm_params : dict
        Extra arguments for the distance metric used to compute the RDMs. Refer to
        :mod:`scipy.spatial.distance` for a list of all other metrics and their
        arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the RDMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    ignore_nan : bool
        Whether to treat NaN's as missing values and ignore them when computing the
        distance metric. Defaults to ``False``.

        .. versionadded:: 0.8
    y : ndarray of int, shape (n_items,) | None
        (Deprecated) For each item, a number indicating the class to which the item
        belongs.  When ``None``, each item is assumed to belong to a different class.
        Defaults to ``None``.
    labels_rdm_model: list | None
        For each row in ``rdm_model``, a label that identifies the item to which it
        corresponds. This is used in combination with ``labels_X`` to align the data and
        model RDMs before comparing them. Each row should have a unique label. Labels
        may be of any python type that can be compared with ``==`` (int, float, string,
        tuple, etc). By default (``None``), the integers ``0:n_rows`` are used as
        labels.

        .. versionadded:: 0.10
    labels_X : list | None
        For each element in ``X`` (=the first dimension), a label that identifies the
        item to which it corresponds. This is used in combination with
        ``labels_rdm_model`` to align the data and model RDMs before comparing them.
        Multiple elements in ``X`` may correspond to the same item, in which case they
        should have the same label and will be averaged when computing the data RDM.
        Labels may be of any python type that can be compared with ``==`` (int, float,
        string, tuple, etc). By default (``None``), the integers ``0:len(X)`` are used
        as labels.

        .. versionadded:: 0.10
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance metric.
        Folds are created based on the ``y`` parameter. Specify ``None`` to use the
        maximum number of folds possible, given the data. Alternatively, you can pass a
        Scikit-Learn cross validator object (e.g. ``sklearn.model_selection.KFold``) to
        assert fine-grained control over how folds are created.
        Defaults to 1 (no cross-validation).
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to use all
        available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need the tqdm
        python module installed. Defaults to False.

    Returns
    -------
    rsa_vals : ndarray, shape ([n_series,] [n_times,] [n_model_rdms])
        The RSA value for each searchlight patch. When ``spatial_radius`` is set to
        ``None``, there will only be no ``n_series`` dimension. When ``temporal_radius``
        is set to ``None``, there will be no time dimension. When multiple models have
        been supplied, the last dimension will contain RSA results for each model.

    See Also
    --------
    searchlight
    compute_rdm
    rdm_array

    """
    if patches is None:
        patches = searchlight(X.shape)  # One big searchlight patch

    # Create folds for cross-validated RDM metrics
    X = create_folds(X, y, n_folds)
    # The data is now folds x items x n_series x n_times

    if isinstance(rdm_model, list):
        rdm_model = [_ensure_condensed(rdm, "rdm_model") for rdm in rdm_model]
    else:
        rdm_model = [_ensure_condensed(rdm_model, "rdm_model")]

    if ignore_nan:
        masks = [~np.isnan(rdm) for rdm in rdm_model]
    else:
        masks = [slice(None)] * len(rdm_model)

    if verbose:
        from tqdm import tqdm

        shape = getattr(patches, "shape", (-1,))
        patches = tqdm(patches, unit="patch")
        try:
            setattr(patches, "shape", shape)
        except AttributeError:
            pass

    def rsa_single_patch(patch):
        """Compute RSA for a single searchlight patch."""
        if len(X) == 1:  # Check number of folds
            # No cross-validation
            rdm_data = compute_rdm(X[0][patch], data_rdm_metric, **data_rdm_params)
        else:
            # Use cross-validation
            rdm_data = compute_rdm_cv(
                X[(slice(None),) + patch], data_rdm_metric, **data_rdm_params
            )
        if ignore_nan:
            data_mask = ~np.isnan(rdm_data)
            patch_masks = [m & data_mask for m in masks]
        else:
            patch_masks = masks
        return _rsa_single_rdm(rdm_data, rdm_model, rsa_metric, patch_masks)

    # Call RSA multiple times in parallel for each searchlight patch
    data = Parallel(n_jobs=n_jobs)(
        delayed(rsa_single_patch)(patch) for patch in patches
    )

    # Figure out the desired dimensions of the resulting array
    dims = getattr(patches, "shape", (-1,))
    if len(rdm_model) > 1:
        dims = dims + (len(rdm_model),)

    return np.array(data).reshape(dims)



__version__ = "0.91"

from .source_level import rsa_stcs, rdm_stcs, rsa_stcs_rois, rsa_nifti, rdm_nifti
from .sensor_level import rsa_evokeds, rsa_epochs, rdm_evokeds, rdm_epochs
from .searchlight import searchlight
from .rsa import rsa, rsa_gen, rsa_array
from .rdm import compute_rdm, compute_rdm_cv, rdm_array, pick_rdm
from .viz import plot_rdms, plot_rdms_topo, plot_roi_map
from .folds import create_folds

# This function is useful to have nearby
from scipy.spatial.distance import squareform



# encoding: utf-8
"""Methods to compute dissimilarity matrices (RDMs)."""

import numpy as np
from joblib import Parallel, delayed
from scipy.spatial import distance

from .folds import create_folds
from .searchlight import searchlight


def compute_rdm(data, metric="correlation", **kwargs):
    """Compute a dissimilarity matrix (RDM).

    Parameters
    ----------
    data : ndarray, shape (n_items, ...)
        For each item, all the features. The first dimension are the items and all other
        dimensions will be flattened and treated as features.
    metric : str | function
        The distance metric to use to compute the RDM. Can be any metric supported by
        :func:`scipy.spatial.distance.pdist`. When a function is specified, it needs to
        take in two vectors and output a single number. See also the ``dist_params``
        parameter to specify and additional parameter for the distance function.
        Defaults to 'correlation'.
    **kwargs : dict, optional
        Extra arguments for the distance metric. Refer to :mod:`scipy.spatial.distance`
        for a list of all other metrics and their arguments.

    Returns
    -------
    rdm : ndarray, shape (n_classes * n_classes-1,)
        The RDM, in condensed form. See :func:`scipy.spatial.distance.squareform`.

    See Also
    --------
    compute_rdm_cv

    """
    X = np.reshape(np.asarray(data), (len(data), -1))
    n_items, n_features = X.shape

    # Be careful with certain metrics
    if n_features == 1 and metric in ["correlation", "cosine"]:
        raise ValueError(
            "There is only a single feature, so "
            "'correlation' and 'cosine' can not be "
            "used as RDM metric. Consider using 'sqeuclidean' "
            "instead."
        )

    return distance.pdist(X, metric, **kwargs)


def compute_rdm_cv(folds, metric="correlation", **kwargs):
    """Compute a dissimilarity matrix (RDM) using cross-validation.

    The distance computation is performed from the average of all-but-one "training"
    folds to the remaining "test" fold. This is repeated with each fold acting as the
    "test" fold once. The resulting distances are averaged and the result used in the
    final RDM.

    Parameters
    ----------
    folds : ndarray, shape (n_folds, n_items, ...)
        For each item, all the features. The first dimension are the folds used for
        cross-validation, items are along the second dimension, and all other dimensions
        will be flattened and treated as features.
    metric : str | function
        The distance metric to use to compute the RDM. Can be any metric supported by
        :func:`scipy.spatial.distance.pdist`. When a function is specified, it needs to
        take in two vectors and output a single number. See also the ``dist_params``
        parameter to specify and additional parameter for the distance function.
        Defaults to 'correlation'.
    **kwargs : dict, optional
        Extra arguments for the distance metric. Refer to :mod:`scipy.spatial.distance`
        for a list of all other metrics and their arguments.

    Returns
    -------
    rdm : ndarray, shape (n_classes * n_classes-1,)
        The cross-validated RDM, in condensed form.
        See :func:`scipy.spatial.distance.squareform`.

    See Also
    --------
    compute_rdm

    """
    X = np.reshape(folds, (folds.shape[0], folds.shape[1], -1))
    n_folds, n_items, n_features = X.shape[:3]

    # Be careful with certain metrics
    if n_features == 1 and metric in ["correlation", "cosine"]:
        raise ValueError(
            "There is only a single feature, so "
            "'correlataion' and 'cosine' can not be "
            "used as RDM metric. Consider using 'sqeuclidean' "
            "instead."
        )

    rdm = np.zeros((n_items * (n_items - 1)) // 2)

    X_mean = X.mean(axis=0)

    # Do cross-validation
    for test_fold in range(n_folds):
        X_test = X[test_fold]
        X_train = X_mean - (X_mean - X_test) / (n_folds - 1)

        dist = distance.cdist(X_train, X_test, metric, **kwargs)
        rdm += dist[np.triu_indices_from(dist, 1)]

    return rdm / n_folds


def _ensure_condensed(rdm, var_name):
    """Convert a RDM to condensed form if needed."""
    if isinstance(rdm, list):
        return [_ensure_condensed(d, var_name) for d in rdm]

    if not isinstance(rdm, np.ndarray):
        raise TypeError(
            "A single RDM should be a NumPy array. "
            "Multiple RDMs should be a list of NumPy arrays."
        )

    if rdm.ndim == 2:
        if rdm.shape[0] != rdm.shape[1]:
            raise ValueError(
                f'Invalid dimensions for "{var_name}" '
                "({rdm.shape}). The RDM should either be a "
                "square matrix, or a one dimensional array when "
                "in condensed form."
            )
        rdm = distance.squareform(rdm)
    elif rdm.ndim != 1:
        raise ValueError(
            f'Invalid dimensions for "{var_name}" ({rdm.shape}). '
            "The RDM should either be a square matrix, or a one "
            "dimensional array when in condensed form."
        )
    return rdm


def _n_items_from_rdm(rdm):
    """Get the number of items, given a RDM."""
    if rdm.ndim == 2:
        return rdm.shape[0]
    elif rdm.ndim == 1:
        return distance.squareform(rdm).shape[0]
    else:
        raise ValueError(f"Wrong number of dimensions for RDM ({rdm.ndim})")


def pick_rdm(rdm, sel):
    """Select items from an RDM.

    Parameters
    ----------
    rdm : ndarray, shape (n, n) | (n * (n - 1) // 2,)
        The RDM to pick items from.
    sel : int | list of int | boolean mask | slice
        The items to pick. These items will be selected from both rows and columns of
        the RDM.

    Returns
    -------
    rdm : ndarray, shape (n, n) | (n * (n - 1) // 2,)
        A new RDM with only the selected items. If the original RDM was in condensed
        form, the returned RDM will be as well.

    """
    if np.isscalar(sel):
        sel = [sel]  # to avoid dropping a dimension
    if rdm.ndim == 2:
        return rdm[sel, :][:, sel]
    elif rdm.ndim == 1:
        return distance.squareform(distance.squareform(rdm)[sel, :][:, sel])
    else:
        raise ValueError(f"Wrong number of dimensions for RDM ({rdm.ndim})")


class rdm_array:
    """Generate RDMs from an array of data, possibly in a searchlight pattern.

    First use :class:`searchlight` to compute the searchlight patches. Then you can use
    this function to compute RDMs for each searchlight patch.

    Parameters
    ----------
    X : ndarray, shape (n_items, n_series, n_times)
        An array containing the data.
    patches : generator of tuples | None
        Searchlight patches as generated by :class:`searchlight`. If ``None``, no
        searchlight is used. Defaults to ``None``.
    dist_metric : str | function
        The distance metric to use to compute the RDMs. Can be any metric supported by
        :func:`scipy.spatial.distance.pdist`. When a function is specified, it needs to
        take in two vectors and output a single number. See also the ``dist_params``
        parameter to specify and additional parameter for the distance function.
        Defaults to 'correlation'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the RDMs. Refer to
        :mod:`scipy.spatial.distance` for a list of all other metrics and their
        arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each item, a number indicating the class to which the item belongs. When
        ``None``, each item is assumed to belong to a different class.
        Defaults to ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance metric.
        Folds are created based on the ``y`` parameter. Specify ``None`` to use the
        maximum number of folds possible, given the data. Alternatively, you can pass a
        Scikit-Learn cross validator object (e.g. ``sklearn.model_selection.KFold``) to
        assert fine-grained control over how folds are created.
        Defaults to 1 (no cross-validation).

    Yields
    ------
    rdm : ndarray, shape (n_patches, n_items * (n_items - 1))
        A RDM (in condensed form) for each searchlight patch.

    Attributes
    ----------
    shape : tuple of int
        Multidimensional shape of the generted RDMs.

        This is useful for re-shaping the result obtained after consuming the this
        generator.

        For a spatio-temporal searchlight:
            Three elements: the number of time-series, number of time samples and length
            of a consensed RDM.
        For a spatial searchlight:
            Two element: the number of time-series and length of a condensed RDM.
        For a temporal searchlight:
            Two elements: the number of time-samples and length of a condensed RDM.
        For no searchlight:
            One element: the length of a condensed RDM.

    See Also
    --------
    rdm
    rsa
    searchlight

    """

    def __init__(
        self,
        X,
        patches=None,
        dist_metric="correlation",
        dist_params=dict(),
        y=None,
        n_folds=1,
        n_jobs=1,
    ):
        if patches is None:
            patches = searchlight(X.shape)

        # Create folds for cross-validated RDM metrics
        self.X = create_folds(X, y, n_folds)
        # The data is now folds x items x n_series x ...

        self.patches = patches
        self.dist_metric = dist_metric
        self.dist_params = dist_params
        self.use_cv = len(self.X) > 1  # More than one fold present

        # Target shape for an array that would hold all of the generated RDMs.
        rdm_length = len(np.triu_indices(self.X.shape[1], k=1)[0])
        self.shape = patches.shape + (rdm_length,)

        self.n_jobs = n_jobs

        # Setup the generator that will be producing the RDMs
        self._generator = iter(self)

    def __iter__(self):
        """Build a new iterator that starts from the beginning."""
        return self._iter_rdms()

    def __next__(self):
        """Generate a RDM for each patch."""
        return next(self._generator)

    def _iter_rdms(self):
        par = Parallel(n_jobs=self.n_jobs, return_as="generator")
        if self.use_cv:
            yield from par(
                delayed(compute_rdm_cv)(
                    self.X[(slice(None),) + patch], self.dist_metric, **self.dist_params
                )
                for patch in self.patches
            )
        else:
            yield from par(
                delayed(compute_rdm)(
                    self.X[0][patch], self.dist_metric, **self.dist_params
                )
                for patch in self.patches
            )

    def __len__(self):
        """Get total number of RDMs that will be generated."""
        return len(self.patches)
